/* -LICENSE-START-
** Copyright (c) 2009 Blackmagic Design
**
** Permission is hereby granted, free of charge, to any person or organization
** obtaining a copy of the software and accompanying documentation covered by
** this license (the "Software") to use, reproduce, display, distribute,
** execute, and transmit the Software, and to prepare derivative works of the
** Software, and to permit third-parties to whom the Software is furnished to
** do so, all subject to the following:
** 
** The copyright notices in the Software and this entire statement, including
** the above license grant, this restriction and the following disclaimer,
** must be included in all copies of the Software, in whole or in part, and
** all derivative works of the Software, unless such copies or derivative
** works are solely in the form of machine-executable object code generated by
** a source language processor.
** 
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
** DEALINGS IN THE SOFTWARE.
** -LICENSE-END-
*/

#include "HDLinkAPI.h"

#include <CoreFoundation/CoreFoundation.h>
#include <iostream>
#include <libgen.h>
#include <libkern/OSAtomic.h>
#include <mach/mach_host.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/select.h>

class HardwareChangeDelegate;

// -- Constants ------------------------------------------------------------ //

enum {
	// LUT cube constants
	kLutCubePoints1D = 17,		// aka "cube dimension"
	kLutCubePointMax1D = kLutCubePoints1D - 1,
	kLutCubePoints2D = kLutCubePoints1D * kLutCubePoints1D,
	kLutCubePoints3D = kLutCubePoints2D * kLutCubePoints1D,
};

// -- Globals -------------------------------------------------------------- //

static IHDLink*					g_hdlink = NULL;
static IHDLinkDiscovery*		g_HDLinkDiscovery = NULL;
static HardwareChangeDelegate*	g_HwDelegate = NULL;
static pthread_cond_t			g_sleepCond;

// -- Protos --------------------------------------------------------------- //

static HRESULT GetHDLinkDevice();
static void Interactive();
static void LoadLUT(bool loadGoodLUT, bool toNVRAM);
static void ConfigToggle();
static void sigfunc(int signum);
static bool HaveInput();
static std::string ModelNameString(IHDLink* hdlink);

// -- Macros --------------------------------------------------------------- //

#define BAIL_IF(cond, fmt, ...)					 \
    if (cond)                                    \
    {                                            \
        fprintf(stderr, "*** %s[%d]: " fmt "\n", \
                __FUNCTION__, __LINE__, ## __VA_ARGS__); \
        goto bail;                               \
    }

// -- Types ---------------------------------------------------------------- //

class HardwareChangeDelegate : public IHDLinkDeviceArrivalCallback, public IHDLinkDeviceRemovalCallback
{
public:
	HardwareChangeDelegate();

	// IUnknown
	virtual HRESULT QueryInterface(REFIID iid, LPVOID *ppv);
	virtual ULONG STDMETHODCALLTYPE AddRef();
	virtual ULONG STDMETHODCALLTYPE Release();

	// IHDLinkDeviceArrivalCallback
	virtual void STDMETHODCALLTYPE HDLinkDeviceArrival(IHDLink *Sender);

	// IHDLinkDeviceRemovalCallback
	virtual void STDMETHODCALLTYPE HDLinkDeviceRemoval(IHDLink *Sender);

private:
	int32_t			m_RefCount;
};

// -- ---------------------------------------------------------------------- //

HardwareChangeDelegate::HardwareChangeDelegate()
	: m_RefCount(1)
{ }
	
HRESULT HardwareChangeDelegate::QueryInterface(REFIID iid, LPVOID *ppv)
{
	CFUUIDBytes iunknown = CFUUIDGetUUIDBytes(IUnknownUUID);

	if (memcmp(&iid, &iunknown, sizeof(REFIID)) == 0 ||
		memcmp(&iid, &IID_IHDLinkDeviceArrivalCallback, sizeof(REFIID)) == 0 ||
		memcmp(&iid, &IID_IHDLinkDeviceRemovalCallback, sizeof(REFIID)) == 0)
	{
		*ppv = this;
		AddRef();
		return S_OK;
	}
	else
	{
		*ppv = NULL;
		return E_NOINTERFACE;
	}
}
	
ULONG HardwareChangeDelegate::AddRef()
{
	return OSAtomicIncrement32(&m_RefCount);
}

ULONG HardwareChangeDelegate::Release()
{
	int newRefValue = OSAtomicDecrement32(&m_RefCount);
	if (newRefValue == 0)
		delete this;
	return newRefValue;
}

void HardwareChangeDelegate::HDLinkDeviceArrival(IHDLink* sender)
{
	HRESULT result = E_FAIL;
	BAIL_IF(g_hdlink != NULL, "Non-NULL IHDLink");

	// we now have reference: must release on exit
	g_hdlink = sender;
	
	result = g_hdlink->InstallDeviceRemovalNotification(this);
	BAIL_IF(result != S_OK, "InstallDeviceRemovalNotification failed 0x%x\n",
			result);

	printf(">>> %s Device Arrived\n", ModelNameString(g_hdlink).c_str());

 bail: ;
}

void HardwareChangeDelegate::HDLinkDeviceRemoval(IHDLink* /*sender*/)
{
	printf(">>> Device Removed\n");

	BAIL_IF(g_hdlink == NULL, "NULL IHDLink");

	g_hdlink->Release();
	g_hdlink = NULL;

 bail: ;
}

void sigfunc(int /*signum*/)
{
	pthread_cond_signal(&g_sleepCond);
}

std::string ModelNameString(IHDLink* hdlink)
{
	char buf[64] = "ERROR";
	CFStringRef cfstring = NULL;
	HRESULT result;
	bool success;

	result = hdlink->GetModelName(&cfstring);
	BAIL_IF(result != S_OK || cfstring == NULL,
			"HDLink GetModelName failed 0x%x\n", result);

	success = CFStringGetCString(cfstring, buf, sizeof(buf), kCFStringEncodingMacRoman);
	BAIL_IF(! success, "GetCString failed\n");

 bail:
	if (cfstring != NULL)
		CFRelease(cfstring);

	return std::string(buf);
}

// NB: currently unused
HRESULT GetHDLinkDeviceWithIterator()
{
	HRESULT result = E_FAIL;
	IHDLinkIterator* hdlinkIterator = NULL;

	result = g_HDLinkDiscovery->CreateHDLinkIterator(&hdlinkIterator);
	BAIL_IF(result != S_OK || hdlinkIterator == NULL,
			"Create HDLinkIterator failed 0x%x\n", result);

	result = hdlinkIterator->Next(&g_hdlink);
	if (result == E_NOINTERFACE)
		goto bail;				// no device found
	BAIL_IF(result != S_OK, "HDLinkIterator Next failed 0x%x\n", result);

	if (g_hdlink != NULL)
		printf("%s Device connected\n", ModelNameString(g_hdlink).c_str());

	result = (g_hdlink == NULL) ? E_FAIL : S_OK;
	
 bail:
	if (hdlinkIterator != NULL)
		hdlinkIterator->Release();

	return result;
}

HRESULT GetHDLinkDeviceWithCallbacks()
{
	HRESULT result = E_FAIL;
	
	g_HDLinkDiscovery = CreateHDLinkDiscoveryInstance();
	BAIL_IF(g_HDLinkDiscovery == NULL,
			"HDLinkDiscovery Create New Instance failed\n");

	g_HwDelegate = new HardwareChangeDelegate();
	BAIL_IF(g_HwDelegate == NULL, "NULL g_HwDelegate\n");

	result = g_HDLinkDiscovery->InstallDeviceAddedNotification(g_HwDelegate);
	BAIL_IF(result != S_OK, "InstallDeviceAddedNotification failed 0x%x\n",
			result);
	// the callback is called during InstallDeviceAddedNotification if
	// any devices are already plugged in.

	result = (g_hdlink == NULL) ? E_FAIL : S_OK;

 bail:
	return result;
}

void LoadLUT(bool loadGoodLUT, bool toNVRAM)
{
	HRESULT				result = S_OK;
	bool				has3dLUT = false;
	LOOKUPTABLE_COLOR*	lut3dTable = NULL;
	const unsigned int	lutSize = kLutCubePoints3D * sizeof(LOOKUPTABLE_COLOR);
	CFStringRef			tableName = loadGoodLUT ? CFSTR("Good LUT") : CFSTR("BAD LUT");

	IHDLinkAttributes*			hdlinkAttribs = NULL;
	IHDLink3DLUTConfiguration*	hdlink3dLutConfig = NULL;

	BAIL_IF(g_hdlink == NULL, "No HDLink connected\n");

	result = g_hdlink->QueryInterface(IID_IHDLinkAttributes, (void**)&hdlinkAttribs);
	BAIL_IF(result != S_OK || hdlinkAttribs == NULL,
			"QueryInterface IHDLinkAttributes failed 0x%x\n", result);

	result = hdlinkAttribs->GetFlag(BMDHDLinkAttribute_Supports3dLUT, &has3dLUT);
	BAIL_IF(result != S_OK, "Attributes GetFlag failed 0x%x\n", result);

	if (! has3dLUT)
	{
		printf("3D Lookup Table not supported by this device\n");
		goto bail;
	}

	result = g_hdlink->QueryInterface(IID_IHDLink3DLUTConfiguration, (void**)&hdlink3dLutConfig);
	BAIL_IF(result != S_OK || hdlink3dLutConfig == NULL,
			"QueryInterface IHDLink3DLUTConfiguration failed 0x%x\n", result);

	lut3dTable = new LOOKUPTABLE_COLOR[lutSize];
	for (int iRed = 0, idx = 0; iRed < kLutCubePoints1D; iRed++)
	{
		for (int iGreen = 0; iGreen < kLutCubePoints1D; iGreen++)
		{
			for (int iBlue = 0; iBlue < kLutCubePoints1D; iBlue++)
			{
				if (loadGoodLUT)
				{
					lut3dTable[idx].red = iRed / float(kLutCubePointMax1D);
					lut3dTable[idx].green = iGreen / float(kLutCubePointMax1D);
					lut3dTable[idx++].blue = iBlue / float(kLutCubePointMax1D);
				}
				else
				{
					lut3dTable[idx].red = (iRed + iGreen) / (2 * float(kLutCubePointMax1D));
					lut3dTable[idx].green = (iGreen + iBlue) / (2 * float(kLutCubePointMax1D));
					lut3dTable[idx++].blue = (iRed + iBlue) / (2 * float(kLutCubePointMax1D));
				}
			}
		}
	}

	result = hdlink3dLutConfig->WriteTable(lut3dTable, kLutCubePoints1D, tableName, toNVRAM);
	BAIL_IF(result != S_OK, "3DLUT WriteTable failed 0x%x\n", result);

	// success
	printf("%s LUT Loaded%s\n",
		   loadGoodLUT ? "Good" : "BAD",
		   toNVRAM ? " in Flash memory" : "");

 bail:
	delete [] lut3dTable;
	if (hdlink3dLutConfig)
		hdlink3dLutConfig->Release();
	if (hdlinkAttribs)
		hdlinkAttribs->Release();
}

void ConfigToggle()
{
	IHDLinkConfiguration* hdlConfig = NULL;
	HRESULT result = S_OK;
	bool bOrig = false;
	bool bNew = false;

	BAIL_IF(g_hdlink == NULL, "No HDLink connected\n");

	// get IConfiguration
	result = g_hdlink->QueryInterface(IID_IHDLinkConfiguration, (void**)&hdlConfig);
	BAIL_IF(result != S_OK, "QueryInterface IHDLinkConfiguration failed\n");

	// query IConfiguration
	result = hdlConfig->GetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, &bOrig);
	BAIL_IF(result != S_OK, "Configuration GetFlag failed 0x%x\n", result);
	
	// toggle setting
	result = hdlConfig->SetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, ! bOrig);
	BAIL_IF(result != S_OK, "Configuration SetFlag failed 0x%x\n", result);

	// query setting again
	result = hdlConfig->GetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, &bNew);
	BAIL_IF(result != S_OK, "Configuration GetFlag failed 0x%x\n", result);

	// check setting
	BAIL_IF(bNew == bOrig, "Configuration Toggle failed!\n");

	// success
	printf("Toggled %s from %s to %s\n", "ScaleIllegalYUV",
		   bOrig ? "true" : "false", bNew ? "true" : "false");
	
 bail:
	if (hdlConfig)
		hdlConfig->Release();
}

bool HaveInput()
{
	struct timeval tv = { 0, 0 };
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(STDIN_FILENO, &fds);

	int result = select(STDIN_FILENO + 1 /*range*/, &fds /*read*/, NULL/*write*/, NULL/*error*/, &tv/*timeout*/);
	BAIL_IF(result < 0, "select failed 0x%x\n", result);

	if (! result)
		return false; // no input yet
	return FD_ISSET(STDIN_FILENO, &fds) != 0;

 bail:
	return false;
}

void Interactive()
{
    printf("Action:\n"
		   "'gv' load good LUT\n"
		   "'bv' load bad LUT\n"
		   "'gn' load good LUT to nonvolatile storage\n"
		   "'bn' load bad LUT to nonvolatile storage\n"
		   "'t'  toggle head/full\n"
		   "'e'  exit\n");

	for (;;)
	{
		if (! HaveInput())
		{
			// service callbacks
			CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0/*wait*/, false/*consume all events*/);
			usleep(100 * 1000);
			continue;
		}

		char str[4];
		bool writeToNv = false;
		int result = scanf("%4s", str);
		BAIL_IF(result != 1, "Expecting 1 value, got %d\n", result);

		switch (str[1])
		{
		case 'v':
			writeToNv = false;
			break;
		case 'n':
			writeToNv = true;
			break;
		}

		char c = str[0];
		switch (c)
		{
		case 'g':
			LoadLUT(true /*loadGoodLUT*/, writeToNv);
			break;
		case 'b':
			LoadLUT(false /*loadGoodLUT*/, writeToNv);
			break;
		case 't':
			ConfigToggle();
			break;
		case 'e':
			return;
		default:
			BAIL_IF(true, "Unrecognised action '%c'\n", c);
		}
	}

 bail: ;
}

// ------------------------------------------------------------------------- //

int main(int argc, char* const argv[])
{
	if (argc <= 1)
	{
		printf("Usage: DemoLutRW [g | b | t | i]\n"
			   "'g' load good LUT\n"
			   "'b' load bad LUT\n"
			   "'t' toggle head/full\n"
			   "'i' interactive menu\n\n");
		return 0;
	}
	
	signal(SIGINT, sigfunc);
	pthread_cond_init(&g_sleepCond, NULL);
	
	if (GetHDLinkDeviceWithCallbacks() != S_OK)
	{
		printf("HDLink Device not connected\n");
		goto bail;
	}

	for (int i = 1; i < argc; i++)
	{
		char c = argv[i][0];
		switch (c)
		{
		case 'g':
			LoadLUT(true /*loadGoodLUT*/, false /*toNVRAM*/);
			break;
		case 'b':
			LoadLUT(false /*loadGoodLUT*/, false /*toNVRAM*/);
			break;
		case 't':
			ConfigToggle();
			break;
		case 'i':
			Interactive();
			break;
		default:
			fprintf(stderr, "Unrecognised option '%c'\n", c);
			break;
		}
	}

 bail:
	if (g_hdlink != NULL)
		g_hdlink->Release();
	if (g_HwDelegate != NULL)
		g_HwDelegate->Release();
	if (g_HDLinkDiscovery != NULL)
		g_HDLinkDiscovery->Release();
	pthread_cond_destroy(&g_sleepCond);

	return 0;
}
