/* -LICENSE-START-
** Copyright (c) 2009 Blackmagic Design
**
** Permission is hereby granted, free of charge, to any person or organization
** obtaining a copy of the software and accompanying documentation covered by
** this license (the "Software") to use, reproduce, display, distribute,
** execute, and transmit the Software, and to prepare derivative works of the
** Software, and to permit third-parties to whom the Software is furnished to
** do so, all subject to the following:
** 
** The copyright notices in the Software and this entire statement, including
** the above license grant, this restriction and the following disclaimer,
** must be included in all copies of the Software, in whole or in part, and
** all derivative works of the Software, unless such copies or derivative
** works are solely in the form of machine-executable object code generated by
** a source language processor.
** 
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
** DEALINGS IN THE SOFTWARE.
** -LICENSE-END-
*/
#include "stdafx.h"

#include "HDLinkAPI.h"

#include <comdef.h> 

// -- Constants ------------------------------------------------------------ //

enum {
	// LUT cube constants
	kLutCubePoints1D = 17,		// aka "cube dimension"
	kLutCubePointMax1D = kLutCubePoints1D - 1,
	kLutCubePoints2D = kLutCubePoints1D * kLutCubePoints1D,
	kLutCubePoints3D = kLutCubePoints2D * kLutCubePoints1D,
};

// -- Globals -------------------------------------------------------------- //

static IHDLink*					g_hdlink = NULL;
static IHDLinkDiscovery*		g_HDLinkDiscovery = NULL;

// -- Protos --------------------------------------------------------------- //

static HRESULT GetHDLinkDevice();
static void Interactive();
static void LoadLUT(bool loadGoodLUT, bool toNVRAM);
static void ConfigToggle();

// -- Macros --------------------------------------------------------------- //

#define BAIL_IF(cond, fmt, ...)					 \
    if (cond)                                    \
    {                                            \
        fprintf(stderr, "*** %s[%d]: " fmt "\n", \
                __FUNCTION__, __LINE__, ## __VA_ARGS__); \
        goto bail;                               \
    }

// ------------------------------------------------------------------------- //

HRESULT GetHDLinkDevice()
{
	HRESULT					result = E_FAIL;
	_bstr_t					hdlinkModelName;
	IHDLinkIterator*		hdlinkIterator = NULL;
	
	g_hdlink = NULL;
	
	result = CoCreateInstance(CLSID_CHDLinkDiscovery, NULL, CLSCTX_ALL, IID_IHDLinkDiscovery, (void**)&g_HDLinkDiscovery);
	BAIL_IF(result != S_OK || g_HDLinkDiscovery == NULL,
			"HDLinkDiscovery Create New Instance failed 0x%x\n", result);

	result = g_HDLinkDiscovery->CreateHDLinkIterator(&hdlinkIterator);
	BAIL_IF(result != S_OK || hdlinkIterator == NULL,
			"HDLinkIterator Create New Instance failed 0x%x\n", result);

	result = hdlinkIterator->Next(&g_hdlink);
	if (result == E_NOINTERFACE)
		goto bail;				// no device found
	BAIL_IF(result != S_OK || g_hdlink == NULL,
			"HDLinkIterator Next failed 0x%x\n", result);

	result = g_hdlink->GetModelName(hdlinkModelName.GetAddress());
	BAIL_IF(result != S_OK, "HDLink GetModelName failed 0x%x\n", result);
	
	// success
	printf("%s Device connected\n", hdlinkModelName.operator const char*());

 bail:
	// Release objects
	if (hdlinkIterator != NULL)
		hdlinkIterator->Release();

	return result;
}

void LoadLUT(bool loadGoodLUT, bool toNVRAM)
{
	HRESULT				result = S_OK;
	BOOL				has3dLUT = false;
	LOOKUPTABLE_COLOR*	lut3dTable = NULL;
	const unsigned int	lutSize = kLutCubePoints3D * sizeof(LOOKUPTABLE_COLOR);
	BSTR				tableName = loadGoodLUT ? OLESTR("Good LUT") : OLESTR("BAD LUT");

	IHDLinkAttributes*			hdlinkAttribs = NULL;
	IHDLink3DLUTConfiguration*	hdlink3dLutConfig = NULL;

	BAIL_IF(g_hdlink == NULL, "No HDLink connected\n");

	result = g_hdlink->QueryInterface(IID_IHDLinkAttributes, (void**)&hdlinkAttribs);
	BAIL_IF(result != S_OK || hdlinkAttribs == NULL,
			"QueryInterface IHDLinkAttributes failed 0x%x\n", result);

	result = hdlinkAttribs->GetFlag(BMDHDLinkAttribute_Supports3dLUT, &has3dLUT);
	BAIL_IF(result != S_OK, "Attributes GetFlag failed 0x%x\n", result);

	if (! has3dLUT)
	{
		printf("3D Lookup Table not supported by this device\n");
		goto bail;
	}

	result = g_hdlink->QueryInterface(IID_IHDLink3DLUTConfiguration, (void**)&hdlink3dLutConfig);
	BAIL_IF(result != S_OK || hdlink3dLutConfig == NULL,
			"QueryInterface IHDLink3DLUTConfiguration failed 0x%x\n", result);

	lut3dTable = new LOOKUPTABLE_COLOR[lutSize];
	for (int iRed = 0, idx = 0; iRed < kLutCubePoints1D; iRed++)
	{
		for (int iGreen = 0; iGreen < kLutCubePoints1D; iGreen++)
		{
			for (int iBlue = 0; iBlue < kLutCubePoints1D; iBlue++)
			{
				if (loadGoodLUT)
				{
					lut3dTable[idx].red = iRed / float(kLutCubePointMax1D);
					lut3dTable[idx].green = iGreen / float(kLutCubePointMax1D);
					lut3dTable[idx++].blue = iBlue / float(kLutCubePointMax1D);
				}
				else
				{
					lut3dTable[idx].red = (iRed + iGreen) / (2 * float(kLutCubePointMax1D));
					lut3dTable[idx].green = (iGreen + iBlue) / (2 * float(kLutCubePointMax1D));
					lut3dTable[idx++].blue = (iRed + iBlue) / (2 * float(kLutCubePointMax1D));
				}
			}
		}
	}

	result = hdlink3dLutConfig->WriteTable(lut3dTable, kLutCubePoints1D, tableName, toNVRAM);
	BAIL_IF(result != S_OK, "3DLUT WriteTable failed 0x%x\n", result);

	// success
	printf("%s LUT Loaded%s\n",
		   loadGoodLUT ? "Good" : "BAD",
		   toNVRAM ? " in Flash memory" : "");

 bail:
	delete [] lut3dTable;
	if (hdlink3dLutConfig)
		hdlink3dLutConfig->Release();
	if (hdlinkAttribs)
		hdlinkAttribs->Release();
}

void ConfigToggle()
{
	IHDLinkConfiguration* hdlConfig = NULL;
	HRESULT result = S_OK;
	BOOL bOrig = false;
	BOOL bNew = false;

	BAIL_IF(g_hdlink == NULL, "No HDLink connected\n");

	// get IConfiguration
	result = g_hdlink->QueryInterface(IID_IHDLinkConfiguration, (void**)&hdlConfig);
	BAIL_IF(result != S_OK, "QueryInterface IHDLinkConfiguration failed\n");

	// query IConfiguration
	result = hdlConfig->GetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, &bOrig);
	BAIL_IF(result != S_OK, "Configuration GetFlag failed 0x%x\n", result);
	
	// toggle setting
	result = hdlConfig->SetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, ! bOrig);
	BAIL_IF(result != S_OK, "Configuration SetFlag failed 0x%x\n", result);

	// query setting again
	result = hdlConfig->GetFlag(BMDHDLinkConfiguration_ScaleIllegalYUV, &bNew);
	BAIL_IF(result != S_OK, "Configuration GetFlag failed 0x%x\n", result);

	// check setting
	BAIL_IF(bNew == bOrig, "Configuration Toggle failed!\n");

	// success
	printf("Toggled %s from %s to %s\n", "ScaleIllegalYUV",
		   bOrig ? "true" : "false", bNew ? "true" : "false");
	
 bail:
	if (hdlConfig)
		hdlConfig->Release();
}

void Interactive()
{
    printf("Action:\n"
		   "'gv' load good LUT\n"
		   "'bv' load bad LUT\n"
		   "'gn' load good LUT to nonvolatile storage\n"
		   "'bn' load bad LUT to nonvolatile storage\n"
		   "'t'  toggle head/full\n"
		   "'e'  exit\n");
	
	for (;;)
	{
		char str[4];
		bool writeToNv = false;
		int result = scanf_s("%4s", str);
		BAIL_IF(result != 1, "Expecting 1 value, got %d\n", result);

		switch (str[1])
		{
		case 'v':
			writeToNv = false;
			break;
		case 'n':
			writeToNv = true;
			break;
		}

		char c = str[0];
		switch (c)
		{
		case 'g':
			LoadLUT(true /*loadGoodLUT*/, writeToNv);
			break;
		case 'b':
			LoadLUT(false /*loadGoodLUT*/, writeToNv);
			break;
		case 't':
			ConfigToggle();
			break;
		case 'e':
			return;
		default:
			BAIL_IF(true, "Unrecognised action '%c'\n", c);
		}
	}

 bail: ;
}

// ------------------------------------------------------------------------- //

int _tmain(int argc, _TCHAR* argv[])
{
	if (argc <= 1)
	{
		printf("Usage: DemoLutRW [g | b | t | i]\n"
			   "'g' load good LUT\n"
			   "'b' load bad LUT\n"
			   "'t' toggle head/full\n"
			   "'i' interactive menu\n\n");
		return 0;
	}
	
	CoInitialize(NULL);
	
	if (GetHDLinkDevice() != S_OK)
	{
		printf("HDLink Device not connected\n");
		goto bail;
	}

	for (int i = 1; i < argc; i++)
	{
		char c = argv[i][0];
		switch (c)
		{
		case 'g':
			LoadLUT(true /*loadGoodLUT*/, false /*toNVRAM*/);
			break;
		case 'b':
			LoadLUT(false /*loadGoodLUT*/, false /*toNVRAM*/);
			break;
		case 't':
			ConfigToggle();
			break;
		case 'i':
			Interactive();
			break;
		default:
			fprintf(stderr, "Unrecognised option '%c'\n", c);
			break;
		}
	}

 bail:
	// Release objects
	if (g_HDLinkDiscovery != NULL)
		g_HDLinkDiscovery->Release();
	CoUninitialize();

	return 0;
}
